@@ -1,4 +1,6 @@
-/* $Id: ifp_routines.c,v 1.24 2004/05/30 15:30:12 yamajun Exp $ */
+/* $Id: ifp_routines.c,v 1.50 2004/11/05 16:44:22 yamajun Exp $ */
+
+static const char rcsid[] = "$Id: ifp_routines.c,v 1.50 2004/11/05 16:44:22 yamajun Exp $";
 
 #include <stdio.h>
 #include <string.h>
@@ -18,8 +20,6 @@
 #define IFP_SEND_DATA			0x18
 #define IFP_LS				0x0f
 #define IFP_LS_NEXT			0x10
-//#define IFP_BULK_TO			0x03
-//#define IFP_BULK_FROM			0x83
 #define IFP_LS_LEAVE			0x11
 #define IFP_GET_FILE_INFO_SIZE		0x0b
 #define IFP_FILE_INFO			0x05
@@ -38,6 +38,10 @@
 #define IFP_02_BATTERY			0x08
 #define IFP_02_STRING			0x00
 #define IFP_02_FIRMWARE			0x03
+#define IFP_02_UPLOAD_FLUSH		0x06
+
+#define IFP_FILE			1
+#define IFP_DIR				2
 
 #define TOUT 1000
 #define FIRMUPDATE_TIMEOUT		5
@@ -71,13 +75,15 @@ int comparefts(const FTSENT **a, const FTSENT **b);
 char *get_basename(const char *path);
 char *ifp_strcasestr(const char *base, const char *target);
 int change_suffix(char *name, const char *newsuffix);
-void remove_double_slash(char *path);
+void normalize_path(char *path, size_t size);
+void remove_double_slash(char *path, size_t size);
 void slash2backslash(char *dest, const char *src, size_t count);
 int ifp_ctl_datasize(unsigned char *control);
 char *dump_control_regs(const char *control, int length);
+int safe_suffix(const char *name);
+// XXX Thats two functions not used.
 int ifp_file_exists(usb_dev_handle *dh, const char *name,int type);
 int get_filepath(char *dest, const char *path);
-int safe_suffix(const char *name);
 
 
 typedef struct dirlists {
@@ -107,7 +113,7 @@ int ifp_put(usb_dev_handle *dh, int argc, char *argv[]) {
 	if (S_ISDIR(status.st_mode)) {
 	    nargv[2] = "/";
 	} else {
-	    nargv[2] = argv[1];
+	    nargv[2] = get_basename(argv[1]);
 	}
     } else {
 	perror("unable to upload file");
@@ -120,6 +126,7 @@ int ifp_put(usb_dev_handle *dh, int argc, char *argv[]) {
 int ifp_upload(usb_dev_handle *dh, int argc, char *argv[]) {
     FILE *fp;
     int retval = 0;
+    char ifp_path[IFP_MAXPATHLEN];
     char path_buf[PATH_MAX];
 
     if (argc < 3) {
@@ -139,23 +146,34 @@ int ifp_upload(usb_dev_handle *dh, int argc, char *argv[]) {
 	    perror(argv[1]);
 	    return -1;
 	} else {
-	    if (ifp_isdir(dh, argv[2])) {
+	    strncpy(ifp_path, argv[2], IFP_MAXPATHLEN - 1);
+	    normalize_path(ifp_path, IFP_MAXPATHLEN);
+
+	    if (ifp_isdir(dh, ifp_path)) {
 		// file to dir
-		//memset(path_buf, 0, sizeof(path_buf));
-		strncpy(path_buf, argv[2], PATH_MAX);
-		strncat(path_buf, "/", 1);
-		strncat(path_buf, get_basename(argv[1]), PATH_MAX - strlen(path_buf) - 1);
-		retval = ifp_upload_file(dh, IS_IFP, path_buf, fp);
+		strncpy(ifp_path, argv[2], IFP_MAXPATHLEN - 2);
+		strncat(ifp_path, "/", 1);
+		strncat(ifp_path, get_basename(argv[1]),
+			IFP_MAXPATHLEN - strlen(ifp_path) - 1);
+		normalize_path(ifp_path, IFP_MAXPATHLEN);
+
+		retval = ifp_upload_file(dh, IS_IFP, ifp_path, fp);
 
 	    } else {
-		retval = ifp_upload_file(dh, IS_IFP, argv[2], fp);  // file to file
+		retval = ifp_upload_file(dh, IS_IFP, ifp_path, fp);  // file to file
 	    }
 	    fclose(fp);
 	}
 	break;
 
     case 2:	// directory
-	retval = ifp_upload_dir(dh, argv[1], argv[2]);   // dir to dir
+	strncpy(path_buf, argv[1], PATH_MAX - 1);
+	normalize_path(path_buf, PATH_MAX);
+
+	strncpy(ifp_path, argv[2], IFP_MAXPATHLEN - 1);
+	normalize_path(ifp_path, IFP_MAXPATHLEN);
+
+	retval = ifp_upload_dir(dh, path_buf, ifp_path);   // dir to dir
 	break;
 
     default:
@@ -171,21 +189,29 @@ int ifp_upload(usb_dev_handle *dh, int argc, char *argv[]) {
     return retval;
 }
 
+// NOTE: "name" need normalize.
 int ifp_upload_file(usb_dev_handle *dh, int flag, const char *name, FILE *f) {
     int i,sum = 0;
     unsigned char buf[IFP_FILE_BUFLEN];
     unsigned char ctl[4];
+    unsigned char ctl8[8];
     int filesize = 0;	// this is a hack, original manager sends file size, this works
     struct stat stat;
+
     fstat(fileno(f),&stat);
     if (stat.st_size > ifp_get_free(dh)) {
+	// not enough space
 	return -2;
+    } else if (strlen(name) >= IFP_MAXPATHLEN) {
+	// Pathname too long
+	return -3;
     }
 
     rewind(f);
     ifp_send_filename(dh,name);
     usb_control_msg(dh,IFP_REQ_TYPE,IFP_FILE_UPLOAD,filesize,0,ctl,sizeof(ctl),TOUT);
     if (!ifp_ok(ctl)) return -1;
+
     while (!feof(f)) {
 	i = fread(buf,1,sizeof(buf),f);
 	//printf("%d bytes read form file...\n",i);
@@ -195,16 +221,19 @@ int ifp_upload_file(usb_dev_handle *dh, int flag, const char *name, FILE *f) {
 	}
 	ifp_send_data(dh,buf,i);
 	usb_control_msg(dh,IFP_REQ_TYPE,IFP_FILE_UPLOAD_APPEND,i,0,ctl,sizeof(ctl),TOUT);
+
 	sum += i;
 	//printf("upload-block-ctl: %d %d %d %d\n",ctl[0],ctl[1],ctl[2],ctl[3],ctl[4]);
     }
     if (flag != IS_MC)
 	printf("%d sent.\n",sum);
-    /* this is a hack - i've lost every last data block, so I send a dummy one */
-    if (isOldIFP) {
-	ifp_send_data(dh,buf,0);
-	usb_control_msg(dh,IFP_REQ_TYPE,IFP_FILE_UPLOAD_APPEND,0,0,ctl,sizeof(ctl),TOUT);
-	//printf("faked-upload-block-ctl: %d %d %d %d\n",ctl[0],ctl[1],ctl[2],ctl[3],ctl[4]);
+
+    while (1) {
+	usb_control_msg(dh, IFP_REQ_TYPE, IFP_02_COMMAND, 0,
+			IFP_02_UPLOAD_FLUSH, ctl8, sizeof(ctl8), TOUT);
+
+	usb_bulk_read(dh, IFP_BULK_FROM, ctl, sizeof(ctl), TOUT);
+	if (ctl[0] != 2) break;
     }
 
     usb_control_msg(dh,IFP_REQ_TYPE,IFP_FILE_LEAVE,0,0,ctl,sizeof(ctl),TOUT);
@@ -233,8 +262,8 @@ int ifp_get(usb_dev_handle *dh, int argc, char *argv[]) {
 }
 
 int ifp_download(usb_dev_handle *dh, int argc, char *argv[]) {
-    FILE *fp;
     int retval = 0;
+    char ifp_path[IFP_MAXPATHLEN];
     char path_buf[PATH_MAX];
 
     if (argc < 3) {
@@ -243,13 +272,16 @@ int ifp_download(usb_dev_handle *dh, int argc, char *argv[]) {
 	return -1;
     }
 
+    strncpy(ifp_path, argv[1], IFP_MAXPATHLEN - 1);
+    normalize_path(ifp_path, IFP_MAXPATHLEN);
+
     switch (filestat(argv[2])) {
     case 0: // "argv[2]" does not exist. create new file.
 	// file to file
-	if (ifp_isdir(dh, argv[1])) {
+	if (ifp_isdir(dh, ifp_path)) {
 	    fprintf(stderr,
 		    "ifp download: Cannot copy %s to %s(%s is a directory).\n",
-		    argv[1], argv[2], argv[1]);
+		    ifp_path, argv[2], ifp_path);
 	    retval = -1;
 	    break;
 	}
@@ -258,22 +290,15 @@ int ifp_download(usb_dev_handle *dh, int argc, char *argv[]) {
 	    break;
 	}
 
-	if ( (fp = fopen(argv[2], "w")) == NULL) {
-	    perror(argv[2]);
-	    retval = -1;
-	} else {
-	    // This function is never call in mc plugin mode.
-	    retval = ifp_download_file(dh, IS_IFP, argv[1], fp);
-	    fclose(fp);
-	}
+	retval = ifp_download_file(dh, IS_IFP, ifp_path, argv[2]);
 	break;
 
     case 1: // regular file
 	// file to file(exist)
-	if (ifp_isdir(dh, argv[1])) {
+	if (ifp_isdir(dh, ifp_path)) {
 	    fprintf(stderr,
 		    "ifp download: Cannot copy %s to %s(%s is a directory).\n",
-		    argv[1], argv[2], argv[1]);
+		    ifp_path, argv[2], ifp_path);
 	    retval = -1;
 	    break;
 	}
@@ -286,13 +311,7 @@ int ifp_download(usb_dev_handle *dh, int argc, char *argv[]) {
 		break;
 	    }
 
-	    if ( (fp = fopen(argv[2], "w")) == NULL) {
-		perror(argv[2]);
-		retval = 1;
-	    } else {
-		retval = ifp_download_file(dh, IS_IFP, argv[1], fp);
-		fclose(fp);
-	    }
+	    retval = ifp_download_file(dh, IS_IFP, ifp_path, argv[2]);
 	    break;
 	default:	/* no */
 	    retval = -1;
@@ -301,28 +320,22 @@ int ifp_download(usb_dev_handle *dh, int argc, char *argv[]) {
 	break;
 
     case 2: // directory
-	if (ifp_isdir(dh, argv[1])) {
+	if (ifp_isdir(dh, ifp_path)) {
 	    // dir to dir
-	    strncpy(path_buf, argv[2], PATH_MAX);
+	    strncpy(path_buf, argv[2], PATH_MAX - 2);
 	    strncat(path_buf, "/", 1);
-	    strncat(path_buf, get_basename(argv[1]), PATH_MAX - strlen(path_buf) - 1);
-	    retval = ifp_download_dir(dh, argv[1], path_buf);
+	    strncat(path_buf, get_basename(ifp_path), PATH_MAX - strlen(path_buf) - 1);
+	    retval = ifp_download_dir(dh, ifp_path, path_buf);
 
 	} else {
 	    // file to dir
-	    strncpy(path_buf, argv[2], PATH_MAX);
+	    strncpy(path_buf, argv[2], PATH_MAX - 2);
 	    strncat(path_buf, "/", 1);
-	    strncat(path_buf, get_basename(argv[1]), PATH_MAX - strlen(path_buf) - 1);
+	    strncat(path_buf, get_basename(ifp_path), PATH_MAX - strlen(path_buf) - 1);
 	    if (!safe_suffix(path_buf)) {
 		break;
 	    }
-	    if ( (fp = fopen(path_buf, "w")) == NULL) {
-		perror(path_buf);
-		retval = -1;
-	    } else {
-		retval = ifp_download_file(dh, IS_IFP, argv[1], fp);
-		fclose(fp);
-	    }
+	    retval = ifp_download_file(dh, IS_IFP, ifp_path, path_buf);
 	}
 	break;
 
@@ -334,33 +347,39 @@ int ifp_download(usb_dev_handle *dh, int argc, char *argv[]) {
     return retval;
 }
 
-int ifp_download_file(usb_dev_handle *dh, int flag, const char *name, FILE *f) {
+// NOTE: "name" need normalize.
+int ifp_download_file(usb_dev_handle *dh, int flag, const char *name, const char *toname) {
+    FILE *fp;
     pid_t pid;
     int i,size,sum = 0;
     unsigned char ctl[4];
     unsigned char ctl8[8];
     unsigned char buf[IFP_FILE_BUFLEN];
     unsigned char typestr[10];
-    char *basename = get_basename(name);
     char *newname;
 
+    if ( (fp = fopen(toname, "w")) == NULL) {
+	perror("ifp_download_file");
+	return -1;
+    }
+
     ifp_send_filename(dh,name);
 
-    //memset(ctl, 0, sizeof(ctl));
-    usb_control_msg(dh,IFP_REQ_TYPE,IFP_FILE_INFO,1,0,ctl,sizeof(ctl),TOUT);
-    if (!ifp_ok(ctl)) {
+    usb_control_msg(dh, IFP_REQ_TYPE, IFP_FILE_INFO, IFP_FILE, 0,
+		    ctl, sizeof(ctl), TOUT);
+
+    if (!ifp_ok(ctl) && flag != IS_MC) {
 	 fprintf(stderr,"ifp download: Could not retrieve file infomation.\n");
 	 return -1;
     }
 
-    //memset(ctl8, 0, sizeof(ctl8));
     usb_control_msg(dh,IFP_REQ_TYPE,IFP_GET_FILE_INFO_SIZE,0,0,ctl8,sizeof(ctl8),TOUT);
     do {
 	usb_control_msg(dh,IFP_REQ_TYPE,IFP_FILE_DOWNLOAD,sizeof(buf),0,ctl,sizeof(ctl),TOUT);
-	size = ctl[0] + ctl[1]*256;
+	size = ctl[0] + (ctl[1]<<8);
 	if (size > 0) {
 	    i = usb_bulk_read(dh,IFP_BULK_FROM,buf,size,TOUT);
-	    fwrite(buf,1,i,f);
+	    fwrite(buf, sizeof(char), i, fp);
 	    sum += i;
 	    if (flag != IS_MC) {
 		printf("%d received\r", sum);
@@ -368,19 +387,24 @@ int ifp_download_file(usb_dev_handle *dh, int flag, const char *name, FILE *f) {
 	    }
 	}
     } while (size > 0);
-    printf("%d received.\n", sum);
+
+    fclose(fp);
+
+    if (flag != IS_MC) {
+	printf("%d received.\n", sum);
+    }
 
     usb_control_msg(dh,IFP_REQ_TYPE,IFP_FILE_LEAVE,0,0,ctl,sizeof(ctl),TOUT);
     if (!ifp_ok(ctl)) return -1;
     
     // convert *.REC file to *.wav/*.mp3
-    if ( (ifp_strcasestr(name, ".REC") - name) == (strlen(name) - 4) ) {
+    if ( (ifp_strcasestr(toname, ".REC") - toname) == (strlen(toname) - 4) ) {
 
-	if ( (newname = malloc(strlen(basename)+1)) == NULL) {
-	    perror("ifp download");
+	if ( (newname = malloc(strlen(toname)+1)) == NULL) {
+	    perror("ifp_download_file");
 	    return -1;
 	}
-	memcpy(newname, basename, strlen(basename)+1);
+	memcpy(newname, toname, strlen(toname)+1);
 
 	ifp_type_string(dh, typestr);
 	if (strstr(typestr, "IFP-1") != NULL) {	// iFP-1XX
@@ -388,10 +412,10 @@ int ifp_download_file(usb_dev_handle *dh, int flag, const char *name, FILE *f) {
 	    change_suffix(newname, ".wav");
 
 	    if ( (pid = fork()) < 0) {
-		perror("ifp download");
+		perror("ifp_download_file");
 		return -1;
 	    } else if (pid == 0) {
-		if (execlp("ifprecconv", "ifprecconv", basename, newname, NULL) == -1)
+		if (execlp("ifprecconv", "ifprecconv", toname, newname, NULL) == -1)
 		    perror("ifprecconv");
 
 	    } else {
@@ -401,8 +425,8 @@ int ifp_download_file(usb_dev_handle *dh, int flag, const char *name, FILE *f) {
 	} else {				// iFP-3XX, iFP-5XX and newer products
 	    // HOGE000.REC -> HOGE000.mp3
 	    change_suffix(newname, ".mp3");
-	    if (rename(basename, newname) == -1) {
-		perror("ifp download");
+	    if (rename(toname, newname) == -1) {
+		perror("ifp_download_file");
 		return -1;
 	    }
 	}
@@ -412,6 +436,7 @@ int ifp_download_file(usb_dev_handle *dh, int flag, const char *name, FILE *f) {
 
 int ifp_rm(usb_dev_handle *dh, int flag, int argc, char *argv[]) {
     int retval = 0;
+    char ifp_path[IFP_MAXPATHLEN];
 
     // for mc support
     // Note: Don't use argv[1] in mc plugin mode!
@@ -431,11 +456,17 @@ int ifp_rm(usb_dev_handle *dh, int flag, int argc, char *argv[]) {
 		printf("usage: rm [-r] file\n");
 		return -1;
 	    } else {
-		retval = ifp_delete_recursive(dh, argv[2]);
+		strncpy(ifp_path, argv[2], IFP_MAXPATHLEN - 1);
+		normalize_path(ifp_path, IFP_MAXPATHLEN);
+
+		retval = ifp_delete_recursive(dh, ifp_path);
 	    }
 
 	} else {
-	    retval = ifp_delete_file(dh, argv[1]);
+	    strncpy(ifp_path, argv[1], IFP_MAXPATHLEN - 1);
+	    normalize_path(ifp_path, IFP_MAXPATHLEN);
+
+	    retval = ifp_delete_file(dh, ifp_path);
 	}
     }
 
@@ -443,66 +474,77 @@ int ifp_rm(usb_dev_handle *dh, int flag, int argc, char *argv[]) {
 }
 
 int ifp_mkdir(usb_dev_handle *dh, int flag, int argc, char *argv[]) {
+    char ifp_path[IFP_MAXPATHLEN];
+
     if (flag == IS_MC) {
 	return ifp_make_dir(dh, argv[2]);
     } else {
-	return ifp_make_dir(dh, argv[1]);
+	strncpy(ifp_path, argv[1], IFP_MAXPATHLEN - 1);
+	normalize_path(ifp_path, IFP_MAXPATHLEN);
+
+	return ifp_make_dir(dh, ifp_path);
     }
 }
 
 int ifp_rmdir(usb_dev_handle *dh, int flag, int argc, char *argv[]) {
+    char ifp_path[IFP_MAXPATHLEN];
+
     if (flag == IS_MC) {
 	return ifp_delete_dir(dh, argv[2]);
     } else {
-	return ifp_delete_dir(dh, argv[1]);
+	strncpy(ifp_path, argv[1], IFP_MAXPATHLEN - 1);
+	normalize_path(ifp_path, IFP_MAXPATHLEN);
+
+	return ifp_delete_dir(dh, ifp_path);
     }
 }
 
 int ifp_list_dir_printf(usb_dev_handle *dh, const char *name, int format, int recurrent) {
     unsigned char ctl[4];
     unsigned char ctl8[8];
-    unsigned char inbuf[IFP_BULK_MAXPATHLEN];
-    unsigned char strbuf[IFP_MAXPATHLEN];
-    unsigned char strbuf2[IFP_MAXPATHLEN];
-    unsigned char strnew[IFP_MAXPATHLEN];
+    char lsbuf[IFP_LSBUFLEN];
+    char ascii_lsbuf[IFP_LSBUFLEN];
+    char filename[IFP_MAXPATHLEN];
+    char ifp_path[IFP_MAXPATHLEN];
     int ret;
     int size;
-    const char *prefix = (strcmp(name,"/") == 0) ? (name+1) : name;
 
-    if (ifp_send_filename(dh,name) < 0) {
-	fprintf(stderr, "ifp ls: Could not send filename: %s.\n", name);
+    strncpy(ifp_path, name, IFP_MAXPATHLEN - 1);
+    normalize_path(ifp_path, IFP_MAXPATHLEN);
+
+    if (ifp_send_filename(dh, ifp_path) < 0) {
+	fprintf(stderr, "ifp ls: Could not send filename: %s.\n",  ifp_path);
 	return -1;
     }
     
     usb_control_msg(dh, IFP_REQ_TYPE,IFP_LS, 0, 0, ctl, sizeof(ctl), TOUT);
     if (!ifp_ok(ctl)) {
-	fprintf(stderr, "ifp ls: %s: No such directory in iFP.\n", name);
+	fprintf(stderr, "ifp ls: %s: No such directory in iFP.\n",  ifp_path);
 	return -1;
     }
 
     while (1) {
-	//memset(ctl, 0, sizeof(ctl));
-	usb_control_msg(dh,IFP_REQ_TYPE,IFP_LS_NEXT,3,0,ctl,sizeof(ctl),TOUT);	// 3=magic !!!???
+	usb_control_msg(dh, IFP_REQ_TYPE, IFP_LS_NEXT, IFP_FILE | IFP_DIR, 0,
+			ctl, sizeof(ctl), TOUT);
 	ret = ifp_ctl_ret(ctl);
 
-	//memset(ctl, 0, sizeof(ctl));
-	//memset(ctl8, 0, sizeof(ctl8));
-	//memset(inbuf, 0, sizeof(inbuf));
-	//memset(strbuf, 0, sizeof(strbuf));
-	//memset(strbuf2, 0, sizeof(strbuf2));
-	//memset(strnew, 0, sizeof(strnew));
-
 	switch (ret) {
 	case 0:
 	    goto LS_END;
 	case 1:
-	    usb_bulk_read(dh,IFP_BULK_FROM,inbuf,sizeof(inbuf),TOUT);
-	    unicode2string(strbuf,inbuf);
-	    strcpy(strbuf2,name);
-	    if (strcmp(name,"/") != 0) strcat(strbuf2,"/");
-	    strcat(strbuf2,strbuf); // path/file
-	    ifp_send_filename(dh,strbuf2);
-	    usb_control_msg(dh,IFP_REQ_TYPE,IFP_FILE_INFO,1,0,ctl,sizeof(ctl),TOUT);
+	    usb_bulk_read(dh, IFP_BULK_FROM, lsbuf, sizeof(lsbuf), TOUT);
+
+	    unicode2string(ascii_lsbuf, lsbuf);
+	    strncpy(filename, ifp_path, IFP_MAXPATHLEN);
+	    if (strcmp(ifp_path, "/") != 0) {
+		strcat(filename, "/");
+	    }
+	    // path/file
+	    strncat(filename, ascii_lsbuf, IFP_MAXPATHLEN-strlen(filename)-1);
+
+	    ifp_send_filename(dh, filename);
+	    usb_control_msg(dh, IFP_REQ_TYPE, IFP_FILE_INFO, IFP_FILE, 0,
+			    ctl, sizeof(ctl), TOUT);
 	    if (!ifp_ok(ctl)) {
 		fprintf(stderr,
 			"ifp ls: File Information command failed. ctl = %s.\n",
@@ -518,28 +560,33 @@ int ifp_list_dir_printf(usb_dev_handle *dh, const char *name, int format, int re
 		 return -1;
 	    }
 
-	    size = ctl8[4] + ctl8[5]*256 + ctl8[6]*256*256 + ctl8[7]*256*256*256;
+	    size = ctl8[4] + (ctl8[5]<<8) + (ctl8[6]<<16) + (ctl8[7]<<24);
 	    if (format) {
-		printf("-rw-rw-rw-   1 root     root  %d Jan  1 1980 %s/%s\n",size,prefix,strbuf);
+		printf("-rw-rw-rw-  1 root  root  %7d Jan  1 1981 %s\n",
+			size, filename);
 	    } else {
-		printf("f %s\t(size %d)\n",strbuf,size);
+		printf("f %s\t(size %d)\n", ascii_lsbuf, size);
 	    }
 	    usb_control_msg(dh,IFP_REQ_TYPE,IFP_FILE_LEAVE,0,0,ctl,sizeof(ctl),TOUT); //??????
 	    break;
 	case 2:
-	    usb_bulk_read(dh,IFP_BULK_FROM,inbuf,sizeof(inbuf),TOUT);
-	    unicode2string(strbuf,inbuf);
+	    usb_bulk_read(dh, IFP_BULK_FROM, lsbuf, sizeof(lsbuf), TOUT);
+	    unicode2string(ascii_lsbuf, lsbuf);
+
+	    strncpy(filename, ifp_path, IFP_MAXPATHLEN);
+	    if (strcmp(filename, "/") != 0) strcat(filename, "/");
+	    // path/file
+	    strncat(filename, ascii_lsbuf, IFP_MAXPATHLEN-strlen(filename)-1);
+
 	    if (format) {
-		printf("drw-rw-rw-   1 root     root  0 Jan  1 1980 %s/%s\n",prefix,strbuf);
+		printf("drw-rw-rw-  1 root  root        0 Jan  1 1980 %s\n",
+			filename);
 	    } else {
-		printf("d %s\n",strbuf);
+		printf("d %s\n", ascii_lsbuf);	//XXX
 	    }
 
 	    if (recurrent) {
-		strcpy(strnew,name);
-		if (strcmp(strnew,"/") != 0) strcat(strnew,"/");
-		strcat(strnew,strbuf);
-		push_dir(strnew);	// push directory with path for later use
+		push_dir(filename); // push directory with path for later use
 	    }
 	    break;
 	default:
@@ -551,8 +598,8 @@ int ifp_list_dir_printf(usb_dev_handle *dh, const char *name, int format, int re
     LS_END:
     usb_control_msg(dh,IFP_REQ_TYPE,IFP_LS_LEAVE,0,0,ctl,sizeof(ctl),TOUT);
     /* now go through the directories on the stack */
-    while (pop_dir(strbuf)) {
-	ifp_list_dir_printf(dh,strbuf,format,recurrent);
+    while (pop_dir(filename)) {
+	ifp_list_dir_printf(dh, filename, format, recurrent);
     }
     return 0;
 }
@@ -561,14 +608,14 @@ int ifp_get_capacity(usb_dev_handle *dh) {
     unsigned char ctl[8];
     usb_control_msg(dh,IFP_REQ_TYPE,IFP_GET_CAPACITY_REQUEST,0,0,(char *)&ctl,sizeof(ctl),TOUT);
     if (!ifp_ok(ctl)) return -1;
-    return ctl[7]*256*256*256 + ctl[6]*256*256 + ctl[5]*256 + ctl[4];
+    return (ctl[7]<<24) + (ctl[6]<<16) + (ctl[5]<<8) + ctl[4];
 }
 
 int ifp_get_free(usb_dev_handle *dh) {
     unsigned char ctl[8];
     usb_control_msg(dh,IFP_REQ_TYPE,IFP_GET_FREE_REQUEST,0,0,ctl,sizeof(ctl),TOUT);
     if (!ifp_ok(ctl)) return -1;
-    return ctl[7]*256*256*256 + ctl[6]*256*256 + ctl[5]*256 + ctl[4];
+    return (ctl[7]<<24) + (ctl[6]<<16) + (ctl[5]<<8) + ctl[4];
 }
 
 /* not tested, not sure: */
@@ -577,9 +624,6 @@ int ifp_battery_condition(usb_dev_handle *dh) {
     unsigned char ctl8[8];
     unsigned char battery[256];
 
-    //memset(ctl8, 0, sizeof(ctl8));
-    //memset(battery, 0, sizeof(battery));
-
     usb_control_msg(dh,IFP_REQ_TYPE,IFP_02_COMMAND,0,IFP_02_BATTERY,ctl8,sizeof(ctl8),TOUT);
     if (!ifp_ok(ctl8)) {
 	fprintf(stderr,
@@ -599,9 +643,6 @@ int ifp_type_string(usb_dev_handle *dh, char *string) {
     unsigned char ctl8[8];
     unsigned char model[256];
 
-    //memset(ctl8, 0, sizeof(ctl8));
-    //memset(model, 0, sizeof(model));
-
     usb_control_msg(dh,IFP_REQ_TYPE,IFP_02_COMMAND,0,IFP_02_STRING,ctl8,sizeof(ctl8),TOUT);
     if (!ifp_ok(ctl8)) {
 	fprintf(stderr, "ifp typestring: Could not get model string from device. ctl8 = %s.\n",
@@ -620,9 +661,6 @@ int ifp_firmware_version(usb_dev_handle *dh) {
     unsigned char ctl8[8];
     unsigned char fversion[256];
 
-    //memset(ctl8, 0, sizeof(ctl8));
-    //memset(fversion, 0, sizeof(fversion));
-
     usb_control_msg(dh,IFP_REQ_TYPE,IFP_02_COMMAND,0,IFP_02_FIRMWARE,ctl8,sizeof(ctl8),TOUT);
     if (!ifp_ok(ctl8)) {
 	fprintf(stderr,
@@ -726,6 +764,11 @@ int ifp_firmware_update(usb_dev_handle *dh, int argc, char *argv[]) {
 
     // close/reopen iFP for update.
     usb_release_interface(dh, dev->config->interface->altsetting->bInterfaceNumber);
+
+//#ifdef linux 
+//    usb_reset(dh);	// Don't use for Mac OS X.
+//#endif
+
     usb_close(dh);
     sleep(FIRMUPDATE_TIMEOUT);	// need this!
 
@@ -733,7 +776,13 @@ int ifp_firmware_update(usb_dev_handle *dh, int argc, char *argv[]) {
 	fprintf(stderr, "Unable to open iFP device.\n");
 	return -1;
     }
-    usb_claim_interface(dh, dev->config->interface->altsetting->bInterfaceNumber);
+
+    if (usb_claim_interface(dh, dev->config->interface->altsetting->bInterfaceNumber)) {
+	fprintf(stderr, "Device is busy.  (I was unable to claim its interface.)\n");
+	usb_close(dh);
+	exit(1);
+    }
+
     if (!ifp_check_connect(dh)) {
 	fprintf(stderr, "Unable to connect to iFP.\n");
 	return -1;
@@ -842,6 +891,7 @@ void info(usb_dev_handle *dh, FILE *f, int head) {
 
 
 /********** internal functions **********/
+// NOTE: "name" need normalize.
 int ifp_upload_dir(usb_dev_handle *dh, const char *from, const char *name) {
     FILE *fp;
     FTS *ftsp;
@@ -849,13 +899,10 @@ int ifp_upload_dir(usb_dev_handle *dh, const char *from, const char *name) {
     size_t basename_pos;
     int fts_options = FTS_LOGICAL;
     int retval = -1;
-    //char *basename = get_basename(from);
     char ifp_path[IFP_MAXPATHLEN];
     char *toname;
     char *argv[2] = {(char*)from, NULL};
-    char *pos;
 
-    //basename_pos = strlen(from) - strlen(basename);
     basename_pos = strlen(from) - strlen(get_basename(from));
 
     if ( (ftsp = fts_open(argv, fts_options, comparefts)) == NULL) {
@@ -865,22 +912,13 @@ int ifp_upload_dir(usb_dev_handle *dh, const char *from, const char *name) {
 
     while ( (ftsentp = fts_read(ftsp)) != NULL) {
 	// "fromdir/file" -> "file"
-	//toname = ftsentp->fts_path + (strlen(from) - strlen(basename));
 	toname = ftsentp->fts_path + basename_pos;
 
 	// "path" "file" -> "path/file"
-	strncpy(ifp_path, name, IFP_MAXPATHLEN);
+	strncpy(ifp_path, name, IFP_MAXPATHLEN - 2);
 	strncat(ifp_path, "/", 1);
 	strncat(ifp_path, toname, IFP_MAXPATHLEN - strlen(ifp_path) - 1);
 
-	/* /path/to//file/// -> /path/to/file/ */
-	remove_double_slash(ifp_path);
-	/* /path/to/file/ -> /path/to/file */
-	for (pos = ifp_path; *pos != '\0'; pos++);
-	if (*(pos-1) == '/') {
-	    *(pos-1) = '\0';
-	}
-
 	switch (ftsentp->fts_info) {
 	case FTS_F:		// normal file
 	    if ( (fp = fopen(ftsentp->fts_path, "r")) == NULL) {
@@ -888,15 +926,18 @@ int ifp_upload_dir(usb_dev_handle *dh, const char *from, const char *name) {
 		goto FTS_END;
 	    }
 
-	    if ((retval = ifp_upload_file(dh, IS_IFP, ifp_path, fp)) < 0) {
+	    // -2: not enough space
+	    if ((retval = ifp_upload_file(dh, IS_IFP, ifp_path, fp)) == -2) {
 		fclose(fp);
 		goto FTS_END;
 	    }
+
 	    fclose(fp);
 
 	    // This is hack. But, it need for upload many files!
 	    // Removed another hack that kills 7XX and 8XX series.
 
+	    //usleep(500000);	// suspend 0.5[sec]
 	    if (isOldIFP) {
 		if (!ifp_check_connect(dh)) {
 		    fprintf(stderr, "Unable to connect to iFP.\n");
@@ -924,17 +965,18 @@ FTS_END:
 }
 
 int ifp_download_dir(usb_dev_handle *dh, const char *fromdir, const char *todir) {
-    FILE *fp;
     int retval = 0;
-    unsigned char inbuf[IFP_BULK_MAXPATHLEN];
-    unsigned char strbuf[IFP_MAXPATHLEN];
+    unsigned char inbuf[IFP_LSBUFLEN];
+    unsigned char strbuf[IFP_LSBUFLEN];
     unsigned char ifp_path[IFP_MAXPATHLEN];
     unsigned char to_path[IFP_MAXPATHLEN];
     unsigned char ctl[4];
 
     if (mkdir(todir, (S_IRWXU | S_IRWXG | S_IRWXO)) == -1) {
-	perror(todir);
-	return -1;
+	if (errno != EEXIST) {
+	    perror(todir);
+	    return -1;
+	}
     }
 
     if (ifp_send_filename(dh, fromdir) < 0)
@@ -947,7 +989,8 @@ int ifp_download_dir(usb_dev_handle *dh, const char *fromdir, const char *todir)
     }
 
     while (1) {
-	usb_control_msg(dh, IFP_REQ_TYPE, IFP_LS_NEXT, 3, 0, ctl, sizeof(ctl), TOUT);
+	usb_control_msg(dh, IFP_REQ_TYPE, IFP_LS_NEXT, IFP_FILE | IFP_DIR, 0,
+			ctl, sizeof(ctl), TOUT);
 	switch (ifp_ctl_ret(ctl)) {
 	case 0:	// end
 	    goto LS_END;
@@ -957,11 +1000,11 @@ int ifp_download_dir(usb_dev_handle *dh, const char *fromdir, const char *todir)
 	    unicode2string(strbuf,inbuf);
 
 	    // "path" "file" -> "path/file"
-	    strncpy(ifp_path, fromdir, IFP_MAXPATHLEN);
+	    strncpy(ifp_path, fromdir, IFP_MAXPATHLEN - 2);
 	    strncat(ifp_path, "/", 1);
 	    strncat(ifp_path, strbuf, IFP_MAXPATHLEN - strlen(ifp_path) - 1);
 
-	    strncpy(to_path, todir, IFP_MAXPATHLEN);
+	    strncpy(to_path, todir, IFP_MAXPATHLEN - 2);
 	    strncat(to_path, "/", 1);
 	    strncat(to_path, strbuf, IFP_MAXPATHLEN - strlen(to_path) - 1);
 
@@ -971,13 +1014,7 @@ int ifp_download_dir(usb_dev_handle *dh, const char *fromdir, const char *todir)
 		    break;
 		}
 
-		if ( (fp = fopen(to_path, "w")) == NULL) {
-		    perror(to_path);
-		    retval = -1;
-		} else {
-		    // This function is never call in mc plugin mode.
-		    retval = ifp_download_file(dh, IS_IFP, ifp_path, fp);
-		}
+		retval = ifp_download_file(dh, IS_IFP, ifp_path, to_path);
 		break;
 	    case 1:	// "to_path" exist
 		printf("WARNING: \"%s\" exist. Do you continue?: ", to_path);
@@ -988,12 +1025,7 @@ int ifp_download_dir(usb_dev_handle *dh, const char *fromdir, const char *todir)
 			break;
 		    }
 
-		    if ( (fp = fopen(to_path, "w")) == NULL) {
-			perror(to_path);
-		    } else {
-			retval = ifp_download_file(dh, IS_IFP, ifp_path, fp);
-			fclose(fp);
-		    }
+		    retval = ifp_download_file(dh, IS_IFP, ifp_path, to_path);
 		    break;
 		default:	/* no */
 		    break;
@@ -1014,7 +1046,7 @@ int ifp_download_dir(usb_dev_handle *dh, const char *fromdir, const char *todir)
 	    unicode2string(strbuf,inbuf);
 
 	    // "path" "file" -> "path/file"
-	    strncpy(ifp_path, fromdir, IFP_MAXPATHLEN);
+	    strncpy(ifp_path, fromdir, IFP_MAXPATHLEN - 2);
 	    strncat(ifp_path, "/", 1);
 	    strncat(ifp_path, get_basename(strbuf), IFP_MAXPATHLEN - strlen(ifp_path) - 1);
 	    retval = push_dir(ifp_path);
@@ -1034,7 +1066,7 @@ LS_END:
     usb_control_msg(dh, IFP_REQ_TYPE, IFP_LS_LEAVE, 0, 0, ctl, sizeof(ctl), TOUT);
     while (pop_dir(strbuf)) {
 	// "path" "file" -> "path/file"
-	strncpy(to_path, todir, IFP_MAXPATHLEN);
+	strncpy(to_path, todir, IFP_MAXPATHLEN - 2);
 	strncat(to_path, "/", 1);
 	strncat(to_path, get_basename(strbuf), IFP_MAXPATHLEN - strlen(to_path) - 1);
 
@@ -1047,11 +1079,11 @@ LS_END:
     return 0;
 }
 
+// NOTE: "name" need normalize.
 int ifp_delete_file(usb_dev_handle *dh, const char *name) {
     unsigned char ctl[4];
 
     ifp_send_filename(dh, name);
-    //memset(ctl, 0, sizeof(ctl));
     usb_control_msg(dh, IFP_REQ_TYPE, IFP_FILE_DELETE, 0, 0, ctl, sizeof(ctl), TOUT);
     if (!ifp_ok(ctl)) return -1;
 
@@ -1059,10 +1091,11 @@ int ifp_delete_file(usb_dev_handle *dh, const char *name) {
 }
 
 // based ifp_list_dir_printf()
+// NOTE: "name" need normalize.
 int ifp_delete_recursive(usb_dev_handle *dh, char *name) {
     int retval = -1;
-    unsigned char inbuf[IFP_BULK_MAXPATHLEN];
-    unsigned char strbuf[IFP_MAXPATHLEN];
+    unsigned char inbuf[IFP_LSBUFLEN];
+    unsigned char strbuf[IFP_LSBUFLEN];
     unsigned char ifp_path[IFP_MAXPATHLEN];
     unsigned char ctl[4];
 
@@ -1073,18 +1106,13 @@ int ifp_delete_recursive(usb_dev_handle *dh, char *name) {
     if (ifp_send_filename(dh, name) < 0)
 	return -1;
 
-    //memset(ctl, 0, sizeof(ctl));
     usb_control_msg(dh, IFP_REQ_TYPE, IFP_LS, 0, 0, ctl, sizeof(ctl), TOUT);
     if (!ifp_ok(ctl))
 	return -1;
 
     while (1) {
-	//memset(ctl, 0, 4);
-	//memset(inbuf, 0, sizeof(inbuf));
-	//memset(strbuf, 0, sizeof(strbuf));
-	//memset(ifp_path, 0, sizeof(ifp_path));
-
-	usb_control_msg(dh, IFP_REQ_TYPE, IFP_LS_NEXT, 3, 0, ctl, sizeof(ctl), TOUT);
+	usb_control_msg(dh, IFP_REQ_TYPE, IFP_LS_NEXT, IFP_FILE | IFP_DIR, 0,
+			ctl, sizeof(ctl), TOUT);
 	switch (ifp_ctl_ret(ctl)) {
 	case 0:	// end
 	    goto LS_END;
@@ -1095,7 +1123,7 @@ int ifp_delete_recursive(usb_dev_handle *dh, char *name) {
 	    unicode2string(strbuf,inbuf);
 
 	    // "path" "file" -> "path/file"
-	    strncpy(ifp_path, name, IFP_MAXPATHLEN);
+	    strncpy(ifp_path, name, IFP_MAXPATHLEN - 2);
 	    strncat(ifp_path, "/", 1);
 	    strncat(ifp_path, strbuf, IFP_MAXPATHLEN - strlen(ifp_path) - 1);
 	    retval = ifp_delete_file(dh, ifp_path);
@@ -1107,7 +1135,7 @@ int ifp_delete_recursive(usb_dev_handle *dh, char *name) {
 	    unicode2string(strbuf,inbuf);
 
 	    // "path" "file" -> "path/file"
-	    strncpy(ifp_path, name, IFP_MAXPATHLEN);
+	    strncpy(ifp_path, name, IFP_MAXPATHLEN - 2);
 	    strncat(ifp_path, "/", 1);
 	    strncat(ifp_path, strbuf, IFP_MAXPATHLEN - strlen(ifp_path) - 1);
 	    retval = push_dir(ifp_path);
@@ -1126,7 +1154,6 @@ int ifp_delete_recursive(usb_dev_handle *dh, char *name) {
     }
 
 LS_END: 
-    //memset(ctl, 0, sizeof(ctl));
     usb_control_msg(dh, IFP_REQ_TYPE, IFP_LS_LEAVE, 0, 0, ctl, sizeof(ctl), TOUT);
     while (pop_dir(strbuf)) {
 	if (ifp_delete_recursive(dh, strbuf) == -1) {
@@ -1137,32 +1164,15 @@ LS_END:
     return ifp_delete_dir(dh, name);
 }
 
+// NOTE: "name" need normalize.
 int ifp_delete_dir(usb_dev_handle *dh, char *name) {
-    int retval;
     unsigned char ctl[4];
-    unsigned char strbuf[IFP_MAXPATHLEN];
-    char *pos;
-
-    //memset(ctl, 0, sizeof(ctl));
-    //memset(strbuf, 0, sizeof(strbuf));
 
     if (strlen(name) <= 0) {
 	printf("usage: rmdir directory\n");
 	return -1;
-
     } 
 
-    // File already created, don't really know whether it is a directory or file
-    retval = ifp_file_exists(dh, name, 2);
-    if (retval == 1) {
-	fprintf(stderr, "%s is not a directory.\n", name);
-	return -1;
-    } else if (retval < 0) {
-	fprintf(stderr, "ERROR: fatal error occurred, unable to stat directory on device %d.\n", retval);
-	retval = -1;
-    }
-
-
     switch (ifp_empty_dir(dh, name)) {
     case 1:
 	fprintf(stderr, "%s: directory not empty\n", name);
@@ -1180,17 +1190,10 @@ int ifp_delete_dir(usb_dev_handle *dh, char *name) {
 
     }
 
-    /* /path/to/file/ -> /path/to/file */
-    strncpy(strbuf, name, IFP_MAXPATHLEN);
-    for (pos = strbuf; *pos != '\0'; pos++);
-    if (*(pos-1) == '/') {
-	*(pos-1) = '\0';
-    }
-
-    ifp_send_filename(dh, strbuf);
+    ifp_send_filename(dh, name);
     usb_control_msg(dh,IFP_REQ_TYPE,IFP_DIR_DELETE,0,0,ctl,sizeof(ctl),TOUT);
     if (!ifp_ok(ctl)) {
-	 fprintf(stderr, "ERROR: Could not remove directory. ctl = %s\n",
+	 fprintf(stderr, "rmdir: Could not remove directory. ctl = %s\n",
 		 dump_control_regs(ctl, 4));
 	 return -1;
     }
@@ -1198,28 +1201,16 @@ int ifp_delete_dir(usb_dev_handle *dh, char *name) {
     return 0;
 }
 
+// NOTE: "name" need normalize.
 int ifp_make_dir(usb_dev_handle *dh, char *name) {
-    int retval;
     unsigned char ctl[4];
 
-    //memset(ctl, 0, sizeof(ctl));
-
     if (strlen(name) <= 0) {
-	printf("usage: mkdir directory\n");
-	return -1;
-    }
-
-    // File already created, don't really know whether it is a directory or file
-    retval = ifp_file_exists(dh, name, 2);
-    if (retval == 1) {
-	fprintf(stderr, "File %s already exists. Skipping...\n", name);
-	return -1;
-    } else if (retval == 2) {
-	fprintf(stderr, "Directory %s already exists. Skipping...\n", name);
-	return -1;
-    } else if (retval < 0) {
-	fprintf(stderr, "ERROR: fatal error occurred, unable to stat directory on device %d.\n", retval);
+	fprintf(stderr, "usage: mkdir directory\n");
 	return -1;
+    } else if (strlen(name) >= IFP_MAXPATHLEN) {
+	fprintf(stderr, "ifp mkdir: Pathname too long.\n");
+	return -2;
     }
 
     ifp_send_filename(dh, name);
@@ -1236,7 +1227,7 @@ int ifp_make_dir(usb_dev_handle *dh, char *name) {
 
 /****** support functions ************/
 int ifp_ctl_ret(unsigned char *control) {
-    return (control[0] + 256*control[1]);
+    return control[0] + (control[1]<<8);
 }
 
 int ifp_ok(unsigned char *control) {
@@ -1286,11 +1277,10 @@ int ifp_empty_dir(usb_dev_handle *dh, const char *name) {
 
     ifp_send_filename(dh,name);
 
-    //memset(ctl, 0, sizeof(ctl));
     usb_control_msg(dh,IFP_REQ_TYPE,IFP_LS,0,0,ctl,sizeof(ctl),TOUT);
     if (!ifp_ok(ctl)) return -1;	// not a dir
-    // XXX 3=magic !!!???
-    usb_control_msg(dh,IFP_REQ_TYPE,IFP_LS_NEXT,3,0,ctl,sizeof(ctl),TOUT);
+    usb_control_msg(dh, IFP_REQ_TYPE, IFP_LS_NEXT, IFP_FILE | IFP_DIR, 0,
+		    ctl, sizeof(ctl), TOUT);
 
     retval = (ifp_ctl_ret(ctl) == 0) ? 0 : 1;
     usb_control_msg(dh,IFP_REQ_TYPE,IFP_LS_LEAVE,0,0,ctl,sizeof(ctl),TOUT);
@@ -1347,10 +1337,11 @@ int comparefts(const FTSENT **a, const FTSENT **b) {
     return strcmp( (*a)->fts_name, (*b)->fts_name);
 }
 
+// NOTE: "path" need normalize.
 char *get_basename(const char *path) {
-    char *start = (char*)path;
+    char *start;
 
-    for (; *path != '\0'; path++) {
+    for (start = (char*)path; *path != '\0'; path++) {
 	if (*path == '/' && *(path+1) != '\0') {
 	    start = (char*)++path;
 	}
@@ -1385,14 +1376,38 @@ int change_suffix(char *name, const char *newsuffix) {
     return -1;
 }
 
-void remove_double_slash(char *path) {
-    char *pos = path;
-    while (*path != '\0') {
-	if (*path == '/' && *(path+1) == '/') {
-	    path++;
-	    continue;
+void normalize_path(char *path, size_t size) {
+    char *pos;
+
+    /* /path/to//file/// -> /path/to/file/ */
+    remove_double_slash(path, size);
+
+    /* /path/to/file/ -> /path/to/file */
+    for (pos = path; *pos != '\0'; pos++)
+	;
+    if (*(pos-1) == '/') {
+	*(pos-1) = '\0';
+    }
+
+}
+
+void remove_double_slash(char *path, size_t size) {
+    int nocopy = 0;
+    int i;
+    char *pos;
+
+    for (i = 0, pos = path; i < size-1 && *path != '\0'; i++) {
+	if (*path == '/') {
+	    if (nocopy == 0) {
+		*pos++ = *path++;
+		nocopy = 1;
+	    } else {
+		path++;
+	    }
+	} else {
+	    *pos++ = *path++;
+	    nocopy = 0;
 	}
-	*pos++ = *path++;
     }
     *pos = '\0';
 }
@@ -1410,7 +1425,7 @@ void slash2backslash(char *dest, const char *src, size_t count) {
 
 
 int ifp_ctl_datasize(unsigned char *control) {
-	return (control[4] + (control[5] * 256));
+    return control[4] + (control[5]<<8);
 }
 
 char *dump_control_regs(const char *control, int length) {
@@ -1445,9 +1460,6 @@ int ifp_file_exists(usb_dev_handle *dh, const char *name, int type) {
     unsigned char path[IFP_MAXPATHLEN];
     unsigned char pathbuf[IFP_MAXPATHLEN];
 
-    memset(path, 0, sizeof(path));
-    memset(pathbuf, 0, sizeof(pathbuf));
-    
     if (name[0] == '/')
 	strcat(pathbuf,name);
     else {
@@ -1468,15 +1480,10 @@ int ifp_file_exists(usb_dev_handle *dh, const char *name, int type) {
     }
     
     while (1) {
-	memset(ctl, 0, sizeof(ctl));
-	// XXX 3=magic !!!???
-	usb_control_msg(dh,IFP_REQ_TYPE,IFP_LS_NEXT,3,0,ctl,sizeof(ctl),TOUT);
+	usb_control_msg(dh, IFP_REQ_TYPE, IFP_LS_NEXT, IFP_FILE | IFP_DIR, 0,
+			ctl, sizeof(ctl), TOUT);
 	ret = ifp_ctl_ret(ctl);
 	
-	memset(ctl, 0, sizeof(ctl));
-	memset(lsbuf, 0, IFP_LSBUFLEN);
-	memset(ascii_lsbuf, 0, IFP_LSBUFLEN);
-	
 	switch (ret) {							
 	case 0:
 	    goto LS_END;
